@@ -1,4 +1,4 @@
-;;; Copyright (C) 2001-2002 by Sam Steingold
+;;; Copyright (C) 2001-2004 by Sam Steingold
 ;;; released under the GNU GPL <http://www.gnu.org/copyleft/gpl.html>
 ;;; as a part of CLISP <http://clisp.cons.org>
 ;;;
@@ -12,14 +12,40 @@
 
 (defun d-mode-translate-word (word)
   "translate the word at point from German into English"
-  (interactive (list (read-string "translate: " (word-at-point))))
+  (interactive (list (read-string "translate: " (thing-at-point 'word))))
   (shell-command (format "wordtrans -d de-en -i %s" word)))
 
+(defun d-mode-find-C-def (sym)
+  (interactive
+   (let ((sym (or (thing-at-point 'symbol)
+                  (save-excursion
+                    (skip-syntax-backward "^w")
+                    (unless (bobp) (backward-char 1))
+                    (thing-at-point 'symbol) ""))))
+     (list (read-string
+            "function name: "
+            (if (string-match ":\\([^:]+\\)$" sym)
+                (match-string 1 sym) sym)))))
+  (let* ((c-name
+          (with-current-buffer (find-file-noselect "constsym.d")
+            (goto-char (point-min))
+            (search-forward (concat "\"" (upcase sym) "\""))
+            (buffer-substring-no-properties
+             (scan-sexps (point) -2) (1- (scan-sexps (point) -1)))))
+         (c-def
+          (with-current-buffer (find-file-noselect "subr.d")
+            (goto-char (point-min))
+            (search-forward (concat "(" c-name ","))
+            (buffer-substring-no-properties
+             (line-beginning-position) (point)))))
+    (message "C-name: %s; c-def: %s" c-name c-def)
+    (tags-search (concat "^" c-def))))
+
 (defun d-mode-current-defun-function ()
   "Return the name of the current function."
   (save-excursion
     (d-mode-beg-of-defun)
-    (cond ((looking-at "LISPFUN")
+    (cond ((looking-at "LISP")
            (search-forward "(")
            (let ((c-name (buffer-substring-no-properties
                           (point) (scan-sexps (point) 1))))
@@ -29,6 +55,9 @@
                (search-forward (concat "LISPSYM(" c-name ","))
                (buffer-substring-no-properties
                 (1+ (point)) (1- (scan-sexps (point) 1))))))
+          ((looking-at "DEFUN")
+           (re-search-forward "([-A-Za-Z]*::?\\([^ ,]+\\)")
+           (match-string 1))
           ((looking-at "\\(local\\|global\\)")
            (search-forward "(") (forward-char -1)
            (let ((beg (scan-sexps (point) -1)))
@@ -46,8 +75,9 @@
   "A valid value for `beginning-of-defun-function' for `d-mode'."
   (re-search-backward
    (eval-when-compile
-    (concat "^" (regexp-opt '("LISPFUN" "local " "global " "#define "
-                              "nonreturning_function" "typedef " "struct ")
+    (concat "^" (regexp-opt '("LISPFUN" "LISPSPECFORM" "local " "global "
+                              "#define " "nonreturning_function"
+                              "typedef " "struct " "DEFUN")
                             t)))
    nil 1))                      ; move to the limit
 
@@ -58,6 +88,7 @@ The point should be on the prototype and the definition should follow."
   (let ((beg (point)))
     (end-of-line 1)
     (delete-region (1- (point)) (- (search-forward "{" nil nil) 2))
+    (forward-char -1) (insert "\n")
     (c-indent-region beg (progn (backward-char 2) (forward-sexp) (point)))))
 
 (defun d-mode-convert-lispfun ()
@@ -66,17 +97,21 @@ The point should be on the prototype and the definition should follow."
   (search-forward "LISPFUN")
   (beginning-of-line)
   (let ((beg (point)))
-    (forward-sexp 2) (insert " {")
-    (search-forward "{") (delete-region (line-beginning-position) (1+ (point)))
+    (forward-sexp 2) (insert "\n{ ") (delete-char 1)
+    (re-search-forward "^ *{")
+    (delete-region (line-beginning-position) (1+ (point)))
     (goto-char beg)
     (c-indent-region beg (progn (forward-sexp 3) (point)))))
 
+(defvar d-comment-start "# ") ; for C++: "// ?"
+(defconst d-comment-start-block (concat "[ \t]*\\(" d-comment-start "\\)"))
+
 (defun d-mode-convert-comment ()
   "Comvert the current comment line from # to /**/"
   (interactive)
   (save-excursion
     (beginning-of-line)
-    (when (search-forward "# " (line-end-position) t)
+    (when (re-search-forward d-comment-start (line-end-position) t)
       (replace-match "/* ") (end-of-line)
       (if (/= ?\\ (char-before)) (insert " */")
         (forward-char -1) (just-one-space) (insert "*/ ")))))
@@ -86,13 +121,22 @@ The point should be on the prototype and the definition should follow."
   (interactive)
   (save-excursion
     (beginning-of-line)
-    (while (looking-at "[ \t]*# ")
+    (while (looking-at d-comment-start-block)
       (forward-line -1))
-    (replace-match "/* ")
+    (replace-match "/* " t t nil 1)
     (forward-line 1)
-    (while (looking-at "[ \t]*# ")
-      (replace-match " ") (forward-line 1))
-    (forward-char -1) (insert " */")))
+    (while (looking-at d-comment-start-block)
+      (replace-match " " t t nil 1) (forward-line 1))
+    (forward-char -1) (skip-chars-backward "\\\\") (just-one-space)
+    (insert "*/ ")))
+
+(defun d-mode-convert-next-comment ()
+  "Convert the next comment appropriately"
+  (interactive)
+  (re-search-forward d-comment-start)
+  (if (progn (beginning-of-line) (looking-at d-comment-start-block))
+      (d-mode-convert-block-comment)
+      (d-mode-convert-comment) (comment-indent)))
 
 (defun d-mode-wrap-do-while ()
   "Wrap this block in do/while(0) [for CPP macros]."
@@ -102,12 +146,14 @@ The point should be on the prototype and the definition should follow."
 
 (defun d-mode-indent-sharp (s-element)
   "Check whether a macro or a comment and indent accordingly."
-  (save-excursion (back-to-indentation) (if (looking-at "# ") 0 [0])))
+  (save-excursion (back-to-indentation)
+                  (if (looking-at d-comment-start) 0 [0])))
 
 (defvar d-font-lock-extra-types
   '(nconc (list "bool" "object" "chart" "[otac]int" "signean" "\\sw+_T"
            "s[aco]int" "[csu]?int[BCDLPQWX0-9]*" "hfint" "fcint" "SPint"
-            "[SU]LONG" "[SU]BYTE" "[DSU]WORD" "[SU]LONGLONG" "RET\\sw*TYPE"
+           "[SU]LONG" "[SU]BYTE" "[DSU]WORD" "[SU]LONGLONG" "RET\\sw*TYPE"
+           "\\sw+_Pseudofun"
            "Values" "SOCKET" "Handle" "stringarg" "FILETIME")
     c-font-lock-extra-types)
   "Extra types to be fontified as such.")
@@ -125,7 +171,7 @@ The point should be on the prototype and the definition should follow."
 (defvar d-extra-keywords
   (eval-when-compile
    (regexp-opt '("var" "local" "global" "true" "false" "NIL" "T" "loop"
-                 "inline" "NULL" "nullobj"
+                 "inline" "NULL" "nullobj" "maygc"
                  "popSTACK" "pushSTACK" "skipSTACK" "skipSTACKop" "STACKop"
                  "dotimespC" "dotimesC" "dotimespL" "dotimesL" "dotimespW"
                  "dotimesW" "nonreturning_function" "return_Values"
@@ -185,11 +231,11 @@ use fontifications, you have to (require 'font-lock) first.  Sorry.
 Beware - this will modify the original C-mode too!"
   (set (make-local-variable 'compile-command)
        (let* ((target (if (eq window-system 'w32) "lisp.exe" "lisp.run"))
-              (prefix "")
+              build-dir
               (make (if (eq window-system 'w32) "nmake" "make"))
               (makefile
-               (cond ((file-readable-p "Makefile") "Makefile")
-                     ((file-readable-p "makefile") "makefile")
+               (cond ((file-readable-p "Makefile") nil)
+                     ((file-readable-p "makefile") nil)
                      ((file-readable-p "makefile-msvc") "makefile-msvc")
                      ((file-readable-p "makefile.msvc") "makefile.msvc")
                      ((file-readable-p "makefile.msvc5") "makefile.msvc5")
@@ -198,10 +244,14 @@ Beware - this will modify the original C-mode too!"
                      ((file-readable-p "makefile-gcc")
                       (setq make "make") "makefile-gcc")
                      ((file-directory-p d-mode-build-dir)
-                      (setq prefix (concat "cd '" d-mode-build-dir "'; "))
+                      (setq build-dir d-mode-build-dir make "make")
                       "Makefile")
-                     (t "<makefile>"))))
-         (concat prefix make " -f " makefile " " target)))
+                     (t nil))))
+         (if build-dir
+             (concat make " -C " build-dir " -f " makefile " " target)
+             (if makefile
+                 (concat make " -f " makefile " " target)
+                 (concat make " " target)))))
   (set (make-local-variable 'add-log-current-defun-function)
        'd-mode-current-defun-function)
   (c-set-offset 'cpp-macro 'd-mode-indent-sharp)
@@ -227,16 +277,73 @@ Beware - this will modify the original C-mode too!"
 
 ;; enable CLISP "# foo" comments
 (modify-syntax-entry ?# ". 1b" d-mode-syntax-table)
-(modify-syntax-entry 32 "- 2b" d-mode-syntax-table) ; space
+(modify-syntax-entry 32 ; space
+                     (if (boundp 'running-xemacs) " 2b" "- 2b")
+                     d-mode-syntax-table)
 (modify-syntax-entry ?\n "> b" d-mode-syntax-table)
 (modify-syntax-entry ?\f "> b" d-mode-syntax-table)
-(modify-syntax-entry ?/ "_ 14" d-mode-syntax-table)
 
 ;; put D buffers along with the C buffers in the menus
-(push '("\\<D\\>" . "C") mouse-buffer-menu-mode-groups)
+(when (boundp 'mouse-buffer-menu-mode-groups)
+  (push '("\\<D\\>" . "C") mouse-buffer-menu-mode-groups))
 
 ;; treat D files like C files for add-log
 (eval-after-load "add-log"
   '(add-to-list 'add-log-c-like-modes 'd-mode))
 
+;; some keybindings
+(define-key d-mode-map (kbd "<f5>") 'd-mode-convert-next-comment)
+
+;; update the dates in headers
+(defvar clisp-home-dir "d:/gnu/clisp/sf/clisp/"
+  "*the location of clisp sources for `clisp-update-dates'")
+(defvar clisp-update-dates-user user-full-name
+  "*default argument for `clisp-update-dates'")
+(defun clisp-update-dates (&optional user)
+  "Update the dates in file header for the user.
+Look at the files that are mentioned in `clisp-home-dir'/src/ChangeLog
+as changed by the `user' and check that this date is in that file's header."
+  (interactive (list (read-from-minibuffer "User: " clisp-update-dates-user)))
+  (setq clisp-update-dates-user user)
+  (message "clisp-update-dates: %s" user)
+  (let* ((c-l (find-file-noselect (expand-file-name "src/ChangeLog"
+                                                    clisp-home-dir)))
+         (year (format-time-string "%Y")) start all bad
+         (re (concat "^" year "-.*" user)))
+    (with-current-buffer c-l
+      (save-excursion
+        (goto-char 0)
+        (while (setq start (re-search-forward re nil t))
+          (let ((end (re-search-forward "^[0-9]")))
+            (goto-char start)
+            (while (re-search-forward "^\t\\* " end t)
+              (setq all (nconc (split-string
+                                (buffer-substring-no-properties
+                                 (point) (re-search-forward
+                                          "[:()]" (line-end-position) t))
+                                "[ ,():]+" t)
+                               all)))))))
+    (setq all (delete-dups all))
+    (message "clisp-update-dates: %d files: %s" (length all) all)
+    (dolist (file all)
+      (let ((buf (find-file-noselect
+                  (expand-file-name
+                   (if (string-match "/" file) file (concat "src/" file))
+                   clisp-home-dir))))
+        (with-current-buffer buf
+          (save-excursion
+            (goto-char 0)
+            (cond ((null (search-forward user nil t))
+                   (message "clisp-update-dates: %s does not mention %s"
+                            file user)
+                   (push file bad))
+                  ((progn (beginning-of-line)
+                          (search-forward year (line-end-position) t))
+                   (message "clisp-update-dates: %s is good!" file)
+                   (kill-buffer buf))
+                  (t (message "clisp-update-dates: %s needs updating" file)
+                     (push file bad)))))))
+    (message "clisp-update-dates: %d files need updating: %s" (length bad) bad)
+    (values bad all)))
+
 (provide 'd-mode)
