@@ -6,24 +6,22 @@
 (mapcar #'fmakunbound '(short-site-name long-site-name))
 
 (defun short-site-name ()
-  (let ((s (or
-            (system::registry "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
+  (let ((s (or (sys::registry "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
                               "RegisteredOrganization")
-            (system::registry "SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
+               (sys::registry "SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
                               "RegisteredOrganization"))))
     (check-type s string)
     s))
 (defun long-site-name ()
-  (let ((s (or
-            (system::registry "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
+  (let ((s (or (sys::registry "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
                               "RegisteredOwner")
-            (system::registry "SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
+               (sys::registry "SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
                               "RegisteredOwner"))))
     (check-type s string)
     s))
 
 (defparameter *editor* "notepad.exe" "The name of the editor.")
-(defun editor-name () (or (sys::getenv "EDITOR") *editor*))
+(defun editor-name () (or (getenv "EDITOR") *editor*))
 
 (defun editor-tempfile ()
   "The temporary file LISP creates for editing."
@@ -31,7 +29,7 @@
 
 (defun edit-file (file)
   "(edit-file file) edits a file."
-  (execute (editor-name) (namestring file t)))
+  (execute (editor-name) (namestring (absolute-pathname file))))
 
 ;; Treat Ctrl-Z in files as whitespace. Some losing middle-age
 ;; editors insist on appending this to files.
@@ -39,8 +37,8 @@
   (set-syntax-from-char #\Code26 #\Space))
 
 (defparameter *load-paths*
-  '(#"C:"                ; erst im Current-Directory von Laufwerk C:
-    #"C:\\CLISP\\...\\") ; dann in allen Directories unterhalb C:\CLISP
+  '(#"C:"               ; Search first in the current directory on drive C:
+    #"C:\\CLISP\\**\\") ; then in all subdirectories under C:\CLISP
   "The list of directories where programs are searched on LOAD etc.
 if device and directory are unspecified:")
 
@@ -50,14 +48,21 @@ if device and directory are unspecified:")
 ;; understand CYGWIN pathnames
 (setq *device-prefix* "cygdrive")
 
+;; This perhaps makes pathname parsing more intuitive:
+;;  ".clisprc" --> #S(pathname :name ".clisprc" :type nil)
+(setq *parse-namestring-dot-file* :name)
+
+;; which browser do you use? (see `*browsers*' in clhs.lisp)
+;; (setq *browser* :mozilla-remote)
+
 ;; Common Lisp HyperSpec access
 (defvar *clhs-root-default*)
 (defun clhs-root ()
   "This returns the root URL for the Common Lisp HyperSpec.
 You can set the environment variable `CLHSROOT' or redefine this function
 in ~/.clisprc.  On win32 you can also use the Registry."
-  (or (sys::getenv "CLHSROOT")
-      (let ((s (system::registry "SOFTWARE\\GNU\\CLISP" "CLHSROOT")))
+  (or (getenv "CLHSROOT")
+      (let ((s (sys::registry "SOFTWARE\\GNU\\CLISP" "CLHSROOT")))
         (check-type s (or null string))
         s)
       *clhs-root-default*))
