@@ -1,5 +1,6 @@
 ;;; CLISP Compiler Macros
 ;;; Sam Steingold 2001-05-09
+;;; Bruno Haible 2005
 ;;; CLHS 3.2.2.1 http://www.lisp.org/HyperSpec/Body/sec_3-2-2-1.html
 
 (in-package "SYSTEM")
@@ -43,45 +44,45 @@
 
 (defun compiler-macro-function (name &optional environment)
   (declare (ignore environment))
-  (setq name (check-function-name 'compiler-macro-function name))
+  (setq name (check-function-name name 'compiler-macro-function))
   (if (symbolp name) (get name 'compiler-macro)
       (get (second name) 'compiler-macro-setf)))
 
-(defun (setf compiler-macro-function) (newf name &optional environment)
+(defun (setf compiler-macro-function) (newf name &optional environment) ; ABI
   (declare (ignore environment))
-  (setq name (check-function-name '(setf compiler-macro-function) name))
+  (setq name (check-function-name name '(setf compiler-macro-function)))
   (if (symbolp name) (setf (get name 'compiler-macro) newf)
       (setf (get (second name) 'compiler-macro-setf) newf)))
 
-;; (proclaim '(inline function-form-funform simple-function-form-p))
+;; (proclaim '(inline function-form-p simple-function-form-p))
 
-;; check whether the form is (FUNCTION fun-form) and return the fun-form
-(defun function-form-funform (form)
+;; Test whether the form is (FUNCTION ...).
+(defun function-form-p (form)
   (and (consp form) (eq (car form) 'FUNCTION)
-       (consp (cdr form)) (null (cddr form))
-       (second form)))
+       (consp (cdr form)) (null (cddr form))))
 
-;; check whether the form is #'symbol
+;; Test whether the form is #'symbol or #'(SETF symbol).
 (defun simple-function-form-p (form)
-  (let ((ff (function-form-funform form)))
-    (and ff (function-name-p ff))))
+  (and (function-form-p form) (function-name-p (second form))))
 
 ;; (funcall (function foo) ...) ==> (foo ...)
-(defun strip-funcall-form (form)
+(defun strip-funcall-form (form) ; ABI
   (if (and (eq (car form) 'funcall) (simple-function-form-p (second form)))
-      (cons (second (second form)) (cddr form))
-      form))
+    (cons (second (second form)) (cddr form))
+    form))
 
-(defmacro define-compiler-macro (&whole form name args &body body)
+(defmacro define-compiler-macro (&whole whole-form
+                                 name args &body body)
   (declare (ignore name args body))
-  (multiple-value-bind (expansion name lambdalist docstring)
-      (sys::make-macro-expansion (cdr form) 'strip-funcall-form)
-    (declare (ignore lambdalist))
+  (multiple-value-bind (expansion expansion-lambdabody name lambdalist docstring)
+      (sys::make-macro-expansion (cdr whole-form) whole-form
+                                 #'function-name-p 'strip-funcall-form)
+    (declare (ignore expansion-lambdabody lambdalist))
     (sys::check-redefinition name 'define-compiler-macro
                              (and (compiler-macro-function name)
                                   "compiler macro"))
     `(EVAL-WHEN (COMPILE LOAD EVAL)
-      ,@(when docstring
-         `((SYSTEM::%SET-DOCUMENTATION ',name 'COMPILER-MACRO ,docstring)))
-      (setf (compiler-macro-function ',name) ,expansion)
-      ',name)))
+       ,@(when docstring
+           `((SYSTEM::%SET-DOCUMENTATION ',name 'COMPILER-MACRO ,docstring)))
+       (SETF (COMPILER-MACRO-FUNCTION ',name) ,expansion)
+       ',name)))
