@@ -1,6 +1,6 @@
 ;;; Foreign function interface for CLISP
 ;;; Bruno Haible 19.2.1995
-;;; Sam Steingold 1998-2003
+;;; Sam Steingold 1998-2005
 
 #+UNICODE
 (progn
@@ -26,7 +26,9 @@
           foreign-address foreign-address-unsigned unsigned-foreign-address
           with-foreign-object with-c-var with-foreign-string
           foreign-allocate allocate-deep allocate-shallow foreign-free
-          foreign-pointer foreign-variable foreign-function))
+          foreign-pointer set-foreign-pointer
+          close-foreign-library memory-as
+          foreign-variable foreign-function))
 
 (eval-when (load compile eval)
   (import (intern "*COUTPUT-FILE*" "COMPILER"))
@@ -34,6 +36,7 @@
   (import (intern "*FFI-MODULE*" "COMPILER"))
   (import (intern "FINALIZE-COUTPUT-FILE" "COMPILER"))
   (import (intern "TEXT" "SYSTEM")) ; messages
+  (import (intern "SYMBOL-TO-KEYWORD" "SYSTEM"))
   (import (intern "CHECK-SYMBOL" "SYSTEM")) ; error checking
   (import (intern "FOREIGN-FUNCTION-IN-ARG-COUNT" "SYSTEM")) ; called by SYS::FUNCTION-SIGNATURE
 )
@@ -71,7 +74,7 @@
                  ;     (or (alphanumericp ch) (eql ch #\_)) ; don't allow #\$
                  ;)
                   (or (char<= #\A c #\Z) (char<= #\a c #\z) (char<= #\0 c #\9)
-                      (char= #\_ c)
+                      (char= #\_ c) (char= #\@ c)
                 ) )
               name
        )
@@ -80,13 +83,15 @@
        (not (gethash name
                      (load-time-value
                       (let* ((reserved-list
-                              '("auto" "break" "case" "char" "continue"
-                                "default" "do" "double" "else" "enum" "extern"
-                                "float" "for" "goto" "if" "int" "long"
-                                "register" "return" "short" "sizeof" "static"
-                                "struct" "switch" "typedef" "union" "unsigned"
-                                "void" "while"))
-                             (reserved-table (make-hash-table :test #'equal)))
+                               '("auto" "break" "case" "char" "continue"
+                                 "default" "do" "double" "else" "enum" "extern"
+                                 "float" "for" "goto" "if" "int" "long"
+                                 "register" "return" "short" "sizeof" "static"
+                                 "struct" "switch" "typedef" "union" "unsigned"
+                                 "void" "while"))
+                             (reserved-table
+                               (make-hash-table :key-type 'string :value-type '(eql t)
+                                                :test #'equal)))
                         (dolist (w reserved-list)
                           (setf (gethash w reserved-table) 'T))
                         reserved-table))))))
@@ -110,7 +115,8 @@
 ;; ============================ C types ============================
 
 ;: The table of C types.
-(defvar *c-type-table* (make-hash-table :test #'eq))
+(defvar *c-type-table*
+  (make-hash-table :key-type 'symbol :test 'stablehash-eq))
 
 ; simple C types
 (dolist (c-type
@@ -124,9 +130,9 @@
 (defun parse-components (typespec)
   (let* ((form-type (first typespec))
          (spec-list
-          (ecase form-type
-            (C-STRUCT (cddr typEspec))
-            (C-UNION (cdr typespec)))))
+           (ecase form-type
+             (C-STRUCT (cddr typEspec))
+             (C-UNION (cdr typespec)))))
     (mapcar (lambda (subspec)
               (unless (and (consp subspec)
                            (eql (length subspec) 2)
@@ -144,45 +150,51 @@
       (VECTOR #'vector)
       (LIST #'list)
       (t (let* ((slots (mapcar #'first (cddr typespec)))
-                (vars (mapcar #'(lambda (x) (gensym (symbol-name x)))
-                              slots))
+                (vars (mapcar #'(lambda (x) (gensym (symbol-name x))) slots))
                 h)
            (eval `(FUNCTION
                    (LAMBDA ,vars
                     (DECLARE (COMPILE))
-                    ,(if (and (setq h (get class 'sys::defstruct-description))
-                              (setq h (svref h 2)))
-                         ;; h is the keyword constructor for the structure
-                         `(,h ,@(mapcan #'(lambda (s v)
-                                            (list (intern (symbol-name s) compiler::*keyword-package*)
-                                                  v))
-                                        slots vars))
-                         ;; no keyword constructor found ->
-                         ;; use CLOS:SLOT-VALUE instead
-                         (let ((ivar (gensym)))
-                           `(LET ((,ivar (CLOS:MAKE-INSTANCE ',class)))
-                             ,@(mapcar #'(lambda (s v)
-                                           `(SETF (CLOS:SLOT-VALUE ,ivar ', s)
-                                             ,v))
-                                       slots vars)
-                             ,ivar)))))))))))
+                    ,(if (and (setq h (get class 'clos::closclass))
+                              (typep h clos::<structure-class>)
+                              (setq h (clos::class-kconstructor h)))
+                       ;; h is the keyword constructor for the structure
+                       `(,h ,@(mapcan #'(lambda (s v)
+                                          (list (symbol-to-keyword s) v))
+                                      slots vars))
+                       ;; no keyword constructor found ->
+                       ;; use CLOS:SLOT-VALUE instead
+                       (let ((ivar (gensym)))
+                         `(LET ((,ivar (CLOS:MAKE-INSTANCE ',class)))
+                           ,@(mapcar #'(lambda (s v)
+                                         `(SETF (CLOS:SLOT-VALUE ,ivar ', s)
+                                           ,v))
+                                     slots vars)
+                           ,ivar)))))))))))
 
 (defmacro with-name/options ((name options name+options) &body body)
-  (let ((no (gensym "NAME+OPTIONS-")))
-    `(let ((,no ,name+options))
-       (multiple-value-bind (,name ,options)
-           (if (consp ,no) (values (first ,no) (rest ,no)) (values ,no nil))
-         (declare (ignorable ,name ,options))
-         ,@body))))
+  (multiple-value-bind (body-rest declarations) (system::parse-body body)
+    (let ((no (gensym "NAME+OPTIONS-")))
+      `(LET ((,no ,name+options))
+         (MULTIPLE-VALUE-BIND (,name ,options)
+             (IF (CONSP ,no) (VALUES (FIRST ,no) (REST ,no)) (VALUES ,no NIL))
+           (DECLARE (IGNORABLE ,name ,options) ,@declarations)
+           ,@body-rest)))))
 
 (defmacro with-defining-c-type ((name c-type len) &body body)
-  `(let ((,c-type (make-array ,len)))
-    (unwind-protect
-         (progn (when ,name (setf (gethash ,name *c-type-table*) ,c-type))
-                ,@body)
-      (when ,name (setf (gethash ,name *c-type-table*) nil)))
-    (when ,name (setf (gethash ,name *c-type-table*) ,c-type))
-    ,c-type))
+  (multiple-value-bind (body-rest declarations) (system::parse-body body)
+    `(LET ((,c-type (MAKE-ARRAY ,len)))
+       ,@(if declarations `((DECLARE ,@declarations)))
+       (UNWIND-PROTECT
+           (PROGN
+             (WHEN ,name
+               (SETF (GETHASH ,name *C-TYPE-TABLE*) ,c-type))
+             ,@body-rest)
+         (WHEN ,name
+           (SETF (GETHASH ,name *C-TYPE-TABLE*) NIL)))
+       (WHEN ,name
+         (SETF (GETHASH ,name *C-TYPE-TABLE*) ,c-type))
+       ,c-type)))
 
 ;; Parse a C type specification.
 ;; If name is non-NIL, it will be assigned to name.
@@ -236,11 +248,12 @@
               (setf (svref c-type 1) (parse-c-type (second typespec)))
               (setf (svref c-type 2) maxdim))))
         (C-FUNCTION
-         (let ((c-type (parse-c-function
-                        (parse-options (rest typespec)
-                                       '(:arguments :return-type :language)
-                                       typespec)
-                        typespec)))
+         (let ((c-type
+                 (parse-c-function
+                   (parse-options (rest typespec)
+                                  '(:arguments :return-type :language)
+                                  typespec)
+                   typespec)))
             (when name (setf (gethash name *c-type-table*) c-type))
             c-type))
         (C-PTR
@@ -253,6 +266,11 @@
           (with-defining-c-type (name c-type 2)
             (setf (svref c-type 0) 'C-PTR-NULL)
             (setf (svref c-type 1) (parse-c-type (second typespec)))))
+        (C-POINTER
+          (unless (eql (length typespec) 2) (invalid typespec))
+          (with-defining-c-type (name c-type 2)
+            (setf (svref c-type 0) 'C-POINTER)
+            (setf (svref c-type 1) (parse-c-type (second typespec)))))
         (C-ARRAY-PTR
           (unless (eql (length typespec) 2) (invalid typespec))
           (with-defining-c-type (name c-type 2)
@@ -260,6 +278,39 @@
             (setf (svref c-type 1) (parse-c-type (second typespec)))))
         (t (invalid typespec))))))
 
+;; Primitive types (symbols) can be safely inlined at compile-time.
+;; Handle (parse-c-type 'uint8) as well as
+;; (def-c-type PGconn c-pointer) (parse-c-type 'pgconn)
+;; In addition, `(c-array uint8 ,(length foo))
+;;  -> (vector 'c-array (parse-c-type 'uint8) (length foo))
+;;  -> (vector 'c-array 'uint8 (length foo))
+;; As a result, execution time is nearly halved.
+(define-compiler-macro parse-c-type (&whole form typespec &optional name)
+  (unless name
+    (cond
+      ((typep typespec '(CONS (EQL QUOTE) (CONS SYMBOL NULL)))
+       (let* ((typespec (second typespec))
+              (internal (gethash typespec *c-type-table* 0)))
+         (when (symbolp internal)
+           (return-from parse-c-type `',internal))))
+      ;; Under Kaz Kylheku's backquote reader, `(c-array x ,l) yields
+      ;;(SYSTEM::BACKQUOTE (C-ARRAY X (SYSTEM::UNQUOTE L)))
+      ;; which is optimized because it is costly, yet a common pattern.
+      ((typep typespec
+              '(CONS (EQL SYS::BACKQUOTE)
+                (CONS
+                 (CONS (MEMBER C-ARRAY C-ARRAY-MAX)
+                  (CONS ATOM ; do not match (SYSTEM::UNQUOTE #) here
+                        (CONS (CONS (EQL SYS::UNQUOTE)
+                                    (CONS * NULL)) NULL))) NULL)))
+       (return-from parse-c-type
+         (let ((typespec (second typespec)))
+           `(VECTOR
+             ',(first typespec)
+             (PARSE-C-TYPE ',(second typespec))
+             (EXT:ETHE UNSIGNED-BYTE ,(second (third typespec)))))))))
+  form)
+
 (defun parse-options (options keywords whole)
   (let ((alist '()))
     (dolist (option options)
@@ -283,8 +334,8 @@
   (append (if (flag-set-p flag ff-language-c) '(:C) '())
           (if (flag-set-p flag ff-language-ansi-c)
               (if (flag-set-p flag ff-language-stdcall)
-                  '(:STDC-STDCALL)
-                  '(:STDC))
+                '(:STDC-STDCALL)
+                '(:STDC))
               '())))
 
 (defun language-to-flag (lang)
@@ -294,7 +345,7 @@
     (:STDC-STDCALL (+ ff-language-ansi-c ff-language-stdcall))))
 
 ;; the default foreign language
-(defvar *foreign-language* nil)
+(defvar *foreign-language* nil) ; ABI
 
 (defmacro default-foreign-language (lang)
   (language-to-flag lang)       ; error checking
@@ -324,8 +375,8 @@
                                          (if (or (eq argtype 'C-STRING)
                                                  (case (ctype-type argtype) ((C-PTR C-PTR-NULL C-ARRAY-PTR) t))
                                                  (eq argmode ':OUT))
-                                             ':ALLOCA
-                                             ':NONE))))
+                                           ':ALLOCA
+                                           ':NONE))))
                         ;; see FOREIGN-CALL-OUT in foreign.d
                         (when (and (or (eq argmode :OUT) (eq argmode :IN-OUT))
                                    (not (eq (ctype-type argtype) 'C-PTR)))
@@ -371,8 +422,8 @@
     (error (TEXT "The name ~S is not a valid C identifier")
            name)))
 
-(defmacro DEF-C-TYPE (&whole whole name typespec)
-  (check-symbol (first whole) name)
+(defmacro DEF-C-TYPE (&whole whole-form name typespec)
+  (setq name (check-symbol name (first whole-form)))
   `(EVAL-WHEN (LOAD COMPILE EVAL)
      (PARSE-C-TYPE ',typespec ',name)
      ',name))
@@ -418,8 +469,8 @@
                                 (list (deparse (svref ctype 1))
                                       (let ((dimensions (subseq ctype 2)))
                                         (if (eql (length dimensions) 1)
-                                            (elt dimensions 0)
-                                            (coerce dimensions 'list)))))
+                                          (elt dimensions 0)
+                                          (coerce dimensions 'list)))))
                                ;; #(c-array-max <c-type> number)
                                (C-ARRAY-MAX
                                 (list (deparse (svref ctype 1))
@@ -445,17 +496,17 @@
                                                         #+AFFI
                                                         ,@(let ((h (logand (ash argflags -8) #xF)))
                                                             (if (not (zerop h))
-                                                               (list (svref *registers* (- h 1)))
-                                                               '())))
+                                                              (list (svref *registers* (- h 1)))
+                                                              '())))
                                                       argspecs))))
                                       (list ':return-type
                                             (deparse (svref ctype 1))
                                             (if (flag-set-p (svref ctype 3) ff-flag-malloc-free) ':MALLOC-FREE ':NONE))
-                                      (cons ':language (flag-to-language
-                                                        (svref ctype 3)))))
+                                      (cons ':language
+                                            (flag-to-language (svref ctype 3)))))
                                ;; #(c-ptr <c-type>), #(c-ptr-null <c-type>)
-                               ;; #(c-array-ptr <c-type>)
-                               ((C-PTR C-PTR-NULL C-ARRAY-PTR)
+                               ;; #(c-array-ptr <c-type>), #(c-pointer <c-type>)
+                               ((C-PTR C-PTR-NULL C-POINTER C-ARRAY-PTR)
                                 (list (deparse (svref ctype 1))))))
                        typespec)))))
       (deparse ctype))))
@@ -470,8 +521,13 @@
 (defstruct ffi-module
   name
   c-name
-  (object-table (make-hash-table :test #'equal))
-  (type-table (make-hash-table :test #'eq))
+  (object-table (make-hash-table :key-type '(or string symbol) :value-type '(cons string fixnum)
+                                 :test 'stablehash-equal :warn-if-needs-rehash-after-gc t))
+  (type-table (make-hash-table :key-type 'symbol :value-type '(or null string)
+                               :test 'stablehash-eq :warn-if-needs-rehash-after-gc t))
+  (init-once '())
+  (init-always '())
+  (fini '())
   (variable-list '())
   (function-list '()))
 (define-symbol-macro *name*
@@ -482,6 +538,12 @@
     (ffi-module-object-table *ffi-module*))
 (define-symbol-macro *type-table*
     (ffi-module-type-table *ffi-module*))
+(define-symbol-macro *init-once*
+    (ffi-module-init-once *ffi-module*))
+(define-symbol-macro *init-always*
+    (ffi-module-init-always *ffi-module*))
+(define-symbol-macro *fini*
+    (ffi-module-fini *ffi-module*))
 (define-symbol-macro *variable-list*
     (ffi-module-variable-list *ffi-module*))
 (define-symbol-macro *function-list*
@@ -501,7 +563,11 @@
 ; Convert a Lisp name to a C name.
 ; (Doesn't really matter how. This must just be a deterministic function.)
 (defun to-c-name (name)
-  (setq name (string name))
+  (setq name (if (and (symbolp name)
+                      (symbol-package name)
+                      (package-case-inverted-p (symbol-package name)))
+               (cs-cl:string name)
+               (string name)))
   (unless (some #'lower-case-p name) (setq name (string-downcase name)))
   (if (c-ident-p name)
     name
@@ -523,7 +589,7 @@
 (defun prepare-c-typedecl (c-type)
   (unless (gethash c-type *type-table*)
     (case (ctype-type c-type)
-      ((c-struct c-union c-array c-array-max)
+      ((C-STRUCT C-UNION C-ARRAY C-ARRAY-MAX)
        (let ((new-typename (symbol-name (gensym "g"))))
          (format *coutput-stream* "~%typedef ~A;~%"
                  (to-c-typedecl c-type new-typename))
@@ -587,7 +653,7 @@
              (c-array-max
               (to-c-typedecl (svref c-type 1)
                              (format nil "(~A)[~D]" name (svref c-type 2))))
-             ((c-ptr c-ptr-null c-array-ptr)
+             ((c-ptr c-ptr-null c-pointer c-array-ptr)
               (to-c-typedecl (svref c-type 1) (format nil "* ~A" name)))
              (c-function
               (to-c-typedecl (svref c-type 1)
@@ -665,9 +731,15 @@
                           *coutput-stream*)))
         (format *coutput-stream* ");~%")))
     (format *coutput-stream*
-            "~%void module__~A__init_function_1 (module_t* module)~%{ }~2%~
-            void module__~A__init_function_2 (module_t* module)~%{~%"
-            *c-name* *c-name*)
+            "~%void module__~A__init_function_1 (module_t* module);~%~
+            ~%void module__~A__init_function_2 (module_t* module);~%~
+            ~%void module__~A__fini_function (module_t* module);~%~
+            ~2%void module__~A__init_function_1 (module_t* module)~%~
+            {~{~%~A~}~%}~2%~
+            void module__~A__init_function_2 (module_t* module)~%~
+            {~{~%~A~}~%"
+            *c-name* *c-name* *c-name*
+            *c-name* *init-once* *c-name* *init-always*)
     (dolist (variable *variable-list*)
       (format *coutput-stream*
               "  register_foreign_variable((void*)&~A,~A,~D,sizeof(~A));~%"
@@ -678,7 +750,10 @@
               "  register_foreign_function((void*)&~A,~A,~D);~%"
               (first function) (to-c-string (first function))
               (svref (second function) 3)))
-    (format *coutput-stream* "}~%")))
+    (format *coutput-stream*
+            "}~2%void module__~A__fini_function (module_t* module)~%~
+            {~{~%~A~}~%}~%"
+            *c-name* *fini*)))
 
 ; Allocate a new object in the module's object_tab.
 (defun new-object (read-only-p initstring)
@@ -705,10 +780,18 @@
 (defmacro C-LINES (format-string &rest args)
   `(EVAL-WHEN (COMPILE)
      (DO-C-LINES ,format-string ,@args)))
-(defun do-c-lines (format-string &rest args)
+(defun do-c-lines (format-string &rest args) ; ABI
   (when (compiler::prepare-coutput-file)
     (prepare-module)
-    (apply #'format *coutput-stream* format-string args)))
+    (etypecase format-string
+      ((or string function)
+       (apply #'format *coutput-stream* format-string args))
+      ((member :init-always :init-once :fini)
+       (let ((code (apply #'format nil args)))
+         (ecase format-string
+           (:init-always (push code *init-always*))
+           (:init-once (push code *init-once*))
+           (:fini (push code *fini*))))))))
 
 ;; ============================ named C variables ============================
 
@@ -717,16 +800,20 @@
     (parse-foreign-name (second name-option))
     (to-c-name lisp-name)))
 
-(defmacro DEF-C-VAR (&whole whole name &rest options)
-  (check-symbol (first whole) name)
-  (let* ((alist (parse-options options '(:name :type :read-only
-                                         :alloc :library)
-                               whole))
+(defmacro DEF-C-VAR (&whole whole-form
+                     name &rest options)
+  (setq name (check-symbol name (first whole-form)))
+  (let* ((alist (parse-options options '(:name :type :read-only :alloc
+                                         :library :documentation)
+                               whole-form))
+         (doc (assoc ':documentation alist))
          (c-name (foreign-name name (assoc ':name alist)))
          (type (second (or (assoc ':type alist)
-                           (sys::error-of-type 'sys::source-program-error
-                                  (TEXT "~S: ~S option missing in ~S")
-                                  'def-c-var ':type whole))))
+                           (sys::error-of-type 'ext:source-program-error
+                             :form whole-form
+                             :detail whole-form
+                             (TEXT "~S: ~S option missing in ~S")
+                             'def-c-var ':type whole-form))))
          (read-only (second (assoc ':read-only alist)))
          (flags (+ (if read-only fv-flag-readonly 0)
                    (let ((alloc (assoc ':alloc alist)))
@@ -758,26 +845,28 @@
          (LOAD-TIME-VALUE
            ,(if library
               `(FFI::FOREIGN-LIBRARY-VARIABLE
-                ',c-name (FFI::FOREIGN-LIBRARY ',library)
+                ',c-name (FFI::FOREIGN-LIBRARY ,library)
                 nil (PARSE-C-TYPE ',type))
               `(LOOKUP-FOREIGN-VARIABLE ',c-name (PARSE-C-TYPE ',type)))))
+       ,@(when doc `((SETF (DOCUMENTATION ',name 'VARIABLE) ',(second doc))))
        (DEFINE-SYMBOL-MACRO ,name
          (FOREIGN-VALUE (LOAD-TIME-VALUE (GET ',name 'FOREIGN-VARIABLE))))
        ',name)))
 
-(defun note-c-var (c-name type flags)
+(defun note-c-var (c-name type flags) ; ABI
   (when (compiler::prepare-coutput-file)
     (prepare-module)
     (push (list c-name (parse-c-type type) flags) *variable-list*)))
 
 (defsetf foreign-value set-foreign-value)
-(defsetf foreign-pointer set-foreign-pointer)
+;(defsetf foreign-pointer set-foreign-pointer) ; no, incompatible with SETF
 (defsetf validp set-validp)
 
 (defmacro with-c-place ((var fvar) &body body)
   (let ((fv (gensym (symbol-name var))))
     `(LET ((,fv ,fvar))
-       (SYMBOL-MACROLET ((,var (FOREIGN-VALUE ,fv))) ,@body))))
+       (SYMBOL-MACROLET ((,var (FOREIGN-VALUE ,fv)))
+         ,@body))))
 
 ;; ============================ Stack allocation ============================
 
@@ -796,39 +885,40 @@
 (defmacro with-foreign-object ((var c-type &optional (init nil init-p))
                                &body body)
   `(EXEC-ON-STACK
-    (LAMBDA (,var) ,@body)
-    (PARSE-C-TYPE ,c-type)
-    . ,(if init-p (list init))))
+     #'(LAMBDA (,var) ,@body)
+     (PARSE-C-TYPE ,c-type)
+     ,@(if init-p `(,init))))
 
 ;; symbol-macro based interface (like DEF-C-VAR)
-; WITH-C-VAR appears as a composition of WITH-FOREIGN-OBJECT and WITH-C-PLACE
+;; WITH-C-VAR appears as a composition of WITH-FOREIGN-OBJECT and WITH-C-PLACE
 (defmacro with-c-var ((var c-type &optional (init nil init-p)) &body body)
   (let ((fv (gensym (symbol-name var))))
     `(EXEC-ON-STACK
-      (LAMBDA (,fv) (SYMBOL-MACROLET ((,var (FOREIGN-VALUE ,fv))) ,@body))
-      (PARSE-C-TYPE ,c-type)
-      . ,(if init-p (list init)))))
+       #'(LAMBDA (,fv)
+           (SYMBOL-MACROLET ((,var (FOREIGN-VALUE ,fv)))
+             ,@body))
+       (PARSE-C-TYPE ,c-type)
+       ,@(if init-p `(,init)))))
 
-(defun exec-with-foreign-string (thunk string
+(defun exec-with-foreign-string (thunk string ; ABI
                                  &key (encoding #+UNICODE custom:*foreign-encoding*
                                                 #-UNICODE custom:*default-file-encoding*)
                                       (null-terminated-p t) (start 0) (end nil))
-  (call-with-foreign-string
-   thunk encoding string start end
-   (if null-terminated-p
-       #-UNICODE 1
-       #+UNICODE (sys::encoding-zeroes encoding) 0)))
+  (call-with-foreign-string thunk encoding string start end
+                            (if null-terminated-p
+                                (sys::encoding-zeroes encoding)
+                                0)))
 
 (defmacro with-foreign-string ((foreign-variable char-count byte-count string
                                 &rest keywords
                                 &key encoding null-terminated-p start end)
                                &body body)
   (declare (ignore encoding null-terminated-p start end)) ; get them via keywords
-  `(exec-with-foreign-string
-    (lambda (,foreign-variable ,char-count ,byte-count) ,@body)
-    ,string .,keywords))
+  `(EXEC-WITH-FOREIGN-STRING
+     #'(LAMBDA (,foreign-variable ,char-count ,byte-count) ,@body)
+     ,string ,@keywords))
 
-;; ============================ heep allocation ============================
+;; ============================ heap allocation ============================
 
 (defmacro allocate-deep (ffi-type initval &rest keywords &key count read-only)
   (declare (ignore count read-only)) ; to be accessed via keywords
@@ -840,6 +930,10 @@
   (declare (ignore count read-only)) ; to be accessed via keywords
   `(foreign-allocate (parse-c-type ,ffi-type) ,@keywords))
 
+;; =========================== low-level interface ===========================
+
+(sys::def-setf-alias memory-as write-memory-as) ; ABI
+
 ;; ============================ named C functions ============================
 
 (defmacro DEF-C-CALL-OUT (name &rest options)
@@ -847,12 +941,14 @@
         'def-c-call-out 'def-call-out)
   `(DEF-CALL-OUT ,name ,@options (:LANGUAGE :STDC)))
 
-(defmacro DEF-CALL-OUT (&whole whole name &rest options)
-  (check-symbol (first whole) name)
-  (let* ((alist (parse-options options '(:name :arguments :return-type
-                                         :language :built-in :library)
-                               whole))
-         (parsed-function (parse-c-function alist whole))
+(defmacro DEF-CALL-OUT (&whole whole-form name &rest options)
+  (setq name (check-symbol name (first whole-form)))
+  (let* ((alist
+          (parse-options options '(:name :arguments :return-type :language
+                                    :built-in :library :documentation)
+                          whole-form))
+         (doc (assoc ':documentation alist))
+         (parsed-function (parse-c-function alist whole-form))
          (signature (argvector-to-signature (svref parsed-function 2)))
          (library (second (assoc :library alist)))
          (c-name (foreign-name name (assoc :name alist))))
@@ -860,19 +956,20 @@
                            alist))
     `(PROGN
        ,(unless library
-          `(EVAL-WHEN (COMPILE) (NOTE-C-FUN ',c-name ',alist ',whole)))
+          `(EVAL-WHEN (COMPILE) (NOTE-C-FUN ',c-name ',alist ',whole-form)))
        (LET ()
          (SYSTEM::REMOVE-OLD-DEFINITIONS ',name)
          (COMPILER::EVAL-WHEN-COMPILE (COMPILER::C-DEFUN ',name ',signature))
+         ,@(when doc `((SETF (DOCUMENTATION ',name 'FUNCTION) ',(second doc))))
          (SYSTEM::%PUTD ',name
             ,(if library
-               `(ffi::foreign-library-function
-                 ',c-name (ffi::foreign-library ',library)
+               `(FFI::FOREIGN-LIBRARY-FUNCTION
+                 ',c-name (FFI::FOREIGN-LIBRARY ,library)
                  nil ,parsed-function)
                `(LOOKUP-FOREIGN-FUNCTION ',c-name ,parsed-function))))
        ',name)))
 
-(defun note-c-fun (c-name alist whole)
+(defun note-c-fun (c-name alist whole) ; ABI
   (when (compiler::prepare-coutput-file)
     (prepare-module)
     (push (list c-name (parse-c-function alist whole)
@@ -880,11 +977,13 @@
           *function-list*)))
 
 #+AFFI
-(defmacro DEF-LIB-CALL-OUT (&whole whole name library &rest options)
-  (check-symbol (first whole) name)
-  (let* ((alist (parse-options options '(:name :offset :arguments :return-type) whole))
+(defmacro DEF-LIB-CALL-OUT (&whole whole-form name library &rest options)
+  (setq name (check-symbol name (first whole-form)))
+  (let* ((alist (parse-options options
+                               '(:name :offset :arguments :return-type)
+                               whole-form))
          (parsed-function
-          (parse-c-function (remove (assoc ':name alist) alist) whole))
+           (parse-c-function (remove (assoc ':name alist) alist) whole-form))
          (signature (argvector-to-signature (svref parsed-function 2)))
          (c-name (foreign-name name (assoc ':name alist)))
          (offset (second (assoc ':offset alist))))
@@ -895,7 +994,7 @@
          (FOREIGN-LIBRARY-FUNCTION ',c-name
            (FOREIGN-LIBRARY ',library)
            ',offset
-           (PARSE-C-FUNCTION ',(remove (assoc ':name alist) alist) ',whole)))
+           (PARSE-C-FUNCTION ',(remove (assoc ':name alist) alist) ',whole-form)))
        ',name)))
 
 (defmacro DEF-C-CALL-IN (name &rest options)
@@ -903,18 +1002,19 @@
         'def-c-call-in 'def-call-in)
   `(DEF-CALL-IN ,name ,@options (:LANGUAGE :STDC)))
 
-(defmacro DEF-CALL-IN (&whole whole name &rest options)
-  (check-symbol (first whole) name)
-  (let* ((alist (parse-options
-                 options '(:name :arguments :return-type :language) whole))
+(defmacro DEF-CALL-IN (&whole whole-form name &rest options)
+  (setq name (check-symbol name (first whole-form)))
+  (let* ((alist (parse-options options
+                               '(:name :arguments :return-type :language)
+                               whole-form))
          (c-name (foreign-name name (assoc ':name alist))))
     (setq alist (remove (assoc ':name alist) alist))
     `(PROGN
        (EVAL-WHEN (COMPILE)
-         (NOTE-C-CALL-IN ',name ',c-name ',alist ',whole))
+         (NOTE-C-CALL-IN ',name ',c-name ',alist ',whole-form))
        ',name)))
 
-(defun note-c-call-in (name c-name alist whole)
+(defun note-c-call-in (name c-name alist whole) ; ABI
   (when (compiler::prepare-coutput-file)
     (prepare-module)
     (let* ((fvd (parse-c-function alist whole))
@@ -925,25 +1025,18 @@
            (argflags (split-c-fun-arglist args 1))
            (argnames (mapcar #'(lambda (argtype) (declare (ignore argtype))
                                  (symbol-name (gensym "g")))
-                             argtypes)))
+                             argtypes))
+           (par-list (mapcar #'to-c-typedecl argtypes argnames))
+           (par-string (format nil "(~{~A~^, ~})" (or par-list '("void"))))
+           (fun+type (to-c-typedecl rettype (format nil "(~A)" c-name))))
       (prepare-c-typedecl rettype)
       ;(mapc #'prepare-c-typedecl argtypes)
-      (format *coutput-stream* "~%extern ~A "
-              (to-c-typedecl rettype (format nil "(~A)" c-name)))
+      ;; prototype + start
+      (format *coutput-stream* "~%~A ~A;~%~A " fun+type par-string fun+type)
       (if (flag-set-p flags ff-language-ansi-c)
-        ; ANSI C parameter declarations
-        (progn
-          (format *coutput-stream* "(")
-          (if argtypes
-            (do ((argtypesr argtypes (cdr argtypesr))
-                 (argnamesr argnames (cdr argnamesr)))
-                ((null argtypesr))
-              (format *coutput-stream* "~A"
-                      (to-c-typedecl (car argtypesr) (car argnamesr)))
-              (when (cdr argtypesr) (format *coutput-stream* ", ")))
-            (format *coutput-stream* "void"))
-          (format *coutput-stream* ")"))
-        ; K&R C parameter declarations
+        ;; ANSI C parameter declarations
+        (format *coutput-stream* "~A" par-string)
+        ;; K&R C parameter declarations
         (progn
           (format *coutput-stream* "(")
           (do ((argnamesr argnames (cdr argnamesr)))
@@ -951,11 +1044,8 @@
             (format *coutput-stream* "~A" (car argnamesr))
             (when (cdr argnamesr) (format *coutput-stream* ", ")))
           (format *coutput-stream* ")")
-          (do ((argtypesr argtypes (cdr argtypesr))
-               (argnamesr argnames (cdr argnamesr)))
-              ((null argtypesr))
-            (format *coutput-stream* "~%  ~A;"
-                    (to-c-typedecl (car argtypesr) (car argnamesr))))))
+          (dolist (par par-list)
+            (format *coutput-stream* "~%  ~A;" par))))
       (format *coutput-stream* "~%{~%  begin_callback();~%")
       (let ((inargcount 0) (outargcount (if (eq rettype 'NIL) 0 1))
             (flag-output (logior ff-flag-out ff-flag-in-out)))
@@ -975,8 +1065,8 @@
   ~A(~A,value1,&retval);~%"
                   (to-c-typedecl rettype "retval")
                   (if (flag-set-p flags ff-flag-malloc-free)
-                      "convert_to_foreign_mallocing"
-                      "convert_to_foreign_nomalloc")
+                    "convert_to_foreign_mallocing"
+                    "convert_to_foreign_nomalloc")
                   (object-to-c-value (pass-object rettype))))
         (let ((outargcount (if (eq rettype 'NIL) 0 1)))
           (mapc #'(lambda (argtype argflag argname)
@@ -986,15 +1076,16 @@
                                'DEF-CALL-IN argtype))
                       (format *coutput-stream* "  ~A~A(~A,~A,~A);~%"
                               (if (eql outargcount 0) ""
-                                  (format nil "if (mv_count >= ~D) "
-                                          (+ outargcount 1)))
+                                (format nil "if (mv_count >= ~D) "
+                                        (+ outargcount 1)))
                               (if (flag-set-p argflag ff-flag-malloc-free)
-                                  "convert_to_foreign_mallocing"
-                                  "convert_to_foreign_nomalloc")
+                                "convert_to_foreign_mallocing"
+                                "convert_to_foreign_nomalloc")
                               (object-to-c-value
-                               (pass-object (svref argtype 1)))
-                              (if (eql outargcount 0) "value1"
-                                  (format nil "mv_space[~D]" outargcount))
+                                (pass-object (svref argtype 1)))
+                              (if (eql outargcount 0)
+                                "value1"
+                                (format nil "mv_space[~D]" outargcount))
                               argname)
                       (incf outargcount)))
                 argtypes argflags argnames))
@@ -1021,9 +1112,9 @@
 (defun foreign-function-in-arg-count (obj)
   (count-inarguments (sys::%record-ref obj 3)))
 
-(defmacro def-c-enum (&whole whole name &rest items)
-  (check-symbol (first whole) name)
-  (let ((forms '()) (ht (make-hash-table))
+(defmacro def-c-enum (&whole whole-form name &rest items)
+  (setq name (check-symbol name (first whole-form)))
+  (let ((forms '()) (ht (make-hash-table :key-type 'fixnum :value-type 'symbol))
         (next-value 0) (this-val 0))
     (dolist (item items)
       (when (consp item)
@@ -1068,57 +1159,76 @@
 ;; (slot (foreign-value x) ...)    --> (foreign-value (%slot x ...))
 ;; (cast (foreign-value x) ...)    --> (foreign-value (%cast x ...))
 ;; (offset (foreign-value x) ...)  --> (foreign-value (%offset x ...))
-(flet ((err (whole)
-         (sys::error-of-type 'sys::source-program-error
+(flet ((err (whole-form reconsed-form)
+         (sys::error-of-type 'ext:source-program-error
+           :form whole-form
+           :detail reconsed-form
            (TEXT "~S is only allowed after ~S: ~S")
-           (first whole) 'FOREIGN-VALUE whole))
+           (first whole-form) 'FOREIGN-VALUE whole-form))
        (foreign-place-p (place type)
          (and (consp place) (eq (first place) type) (eql (length place) 2))))
-  (defmacro element (place &rest indices &environment env)
+
+  (defmacro element (&whole whole-form
+                     place &rest indices &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%ELEMENT ,(second place) ,@indices))
-      (err `(element ,place ,@indices))))
-  (defmacro deref (place &environment env)
+      (err whole-form `(element ,place ,@indices))))
+
+  (defmacro deref (&whole whole-form
+                   place &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%DEREF ,(second place)))
-      (err `(deref ,place))))
-  (defmacro slot (place slotname &environment env)
+      (err whole-form `(deref ,place))))
+
+  (defmacro slot (&whole whole-form
+                  place slotname &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%SLOT ,(second place) ,slotname))
-      (err `(slot ,place ,slotname))))
-  (defmacro cast (place type &environment env)
+      (err whole-form `(slot ,place ,slotname))))
+
+  (defmacro cast (&whole whole-form
+                  place type &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%CAST ,(second place) (PARSE-C-TYPE ,type)))
-      (err `(cast ,place ,type))))
-  (defmacro offset (place offset type &environment env)
+      (err whole-form `(cast ,place ,type))))
+
+  (defmacro offset (&whole whole-form
+                    place offset type &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-VALUE (%OFFSET ,(second place) ,offset (PARSE-C-TYPE ,type)))
-      (err `(offset ,place ,offset ,type))))
+      (err whole-form `(offset ,place ,offset ,type))))
+
   ;; Extract FOREIGN-VARIABLE object underlying the place
-  (defmacro c-var-object (place &environment env)
+  (defmacro c-var-object (&whole whole-form
+                          place &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       (second place)
-      (err `(c-var-object ,place))))
+      (err whole-form `(c-var-object ,place))))
+
   ;; The equivalent of (FOREIGN-ADDRESS fvar) for c-places:
   ;; (c-var-address (foreign-value x)) --> (foreign-address x)
-  (defmacro c-var-address (place &environment env)
+  (defmacro c-var-address (&whole whole-form
+                           place &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(FOREIGN-ADDRESS ,(second place))
-      (err `(c-var-address ,place))))
+      (err whole-form `(c-var-address ,place))))
+
   ;; Similarly for TYPEOF.
   ;; (typeof (foreign-value x)) --> (deparse-c-type (foreign-type x))
-  (defmacro typeof (place &environment env)
+  (defmacro typeof (&whole whole-form
+                    place &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
       `(DEPARSE-C-TYPE (FOREIGN-TYPE ,(second place)))
-      (err `(typeof ,place))))
+      (err whole-form `(typeof ,place))))
+
   ;; Similar tricks are being played for SIZEOF, BITSIZEOF.
   ;; They are macros which work on <c-place>s.
   ;; If the argument is not a <c-place>, they behave like
@@ -1135,6 +1245,7 @@
       (if (foreign-place-p place 'DEPARSE-C-TYPE)
         `(%SIZEOF ,(second place))
         `(%SIZEOF (PARSE-C-TYPE ,place)))))
+
   (defmacro bitsizeof (place &environment env)
     (setq place (macroexpand place env))
     (if (foreign-place-p place 'FOREIGN-VALUE)
