@@ -1,12 +1,15 @@
 (common-lisp:in-package "SYSTEM")
 ;; ----------------------------------------------------------------------------
-(defmacro typecase (keyform &rest typeclauselist)
+(defmacro typecase (&whole whole-form
+                    keyform &rest typeclauselist)
   (let* ((tempvar (gensym))
          (condclauselist nil))
     (do ((typeclauselistr typeclauselist (cdr typeclauselistr)) spec)
         ((atom typeclauselistr))
       (cond ((atom (car typeclauselistr))
              (error-of-type 'source-program-error
+               :form whole-form
+               :detail (car typeclauselistr)
                (TEXT "Invalid clause in ~S: ~S")
                'typecase (car typeclauselistr)))
             ((eq (setq spec (caar typeclauselistr)) T)
@@ -19,19 +22,19 @@
                      condclauselist))))
     `(LET ((,tempvar ,keyform)) (COND ,@(nreverse condclauselist)))))
 ;; ----------------------------------------------------------------------------
-(defun type-error-string ()
+(defun type-error-string () ; ABI
   (TEXT "~A~%The value is: ~S"))
-(defun check-type-error-string (place string typespec)
+(defun check-type-error-string (place string typespec) ; ABI
   (format nil
     (TEXT "The value of ~S should be ~:[of type ~S~;~:*~A~].")
     place string typespec))
-(defun report-one-new-value-string ()
+(defun report-one-new-value-string () ; ABI
   (TEXT "You may input a new value for ~S."))
 (defun report-one-new-value-string-instead ()
   (TEXT "You may input a value to be used instead~@[ of ~S~]."))
-(defun prompt-for-new-value-string ()
-  (TEXT "~%New ~S: "))
-(defmacro check-type (place typespec &optional (string nil))
+(defun prompt-for-new-value-string () ; ABI
+  (concatenate 'string "~&" (TEXT "New ~S: ")))
+(predefmacro check-type (place typespec &optional (string nil))
   (let ((tag1 (gensym))
         (tag2 (gensym)))
     `(TAGBODY
@@ -46,25 +49,24 @@
        (GO ,tag1)
        ,tag2)))
 ;; ----------------------------------------------------------------------------
-(defun report-no-new-value-string ()
+(defun report-no-new-value-string () ; ABI
   (TEXT "Retry"))
-(defun report-new-values-string ()
-  (TEXT "You may input new values for ~S."))
-(defun assert-error-string (test-form)
+(defun report-new-values-string () ; ABI
+  (TEXT "You may input new values for ~@{~S~^, ~}."))
+(defun assert-error-string (test-form) ; ABI
   (format nil
     (TEXT "~S must evaluate to a non-NIL value.")
     test-form))
-(defmacro assert (test-form &optional (place-list nil) (string nil) &rest args)
+(predefmacro assert (test-form &optional (place-list nil) (string nil) &rest args)
   (let ((tag1 (gensym))
         (tag2 (gensym)))
     `(TAGBODY
        ,tag1
        (WHEN ,test-form (GO ,tag2))
-       (CERROR (format nil ,(case (length place-list)
-                                  (0 (REPORT-NO-NEW-VALUE-STRING))
-                                  (1 (REPORT-ONE-NEW-VALUE-STRING))
-                                  (t (REPORT-NEW-VALUES-STRING)))
-                       place-list)
+       (CERROR ,(case (length place-list)
+                  (0 `(REPORT-NO-NEW-VALUE-STRING))
+                  (1 `(APPLY #'FORMAT NIL (REPORT-ONE-NEW-VALUE-STRING) ',place-list))
+                  (t `(APPLY #'FORMAT NIL (REPORT-NEW-VALUES-STRING) ',place-list)))
                ',(or string "~A")
                ,@(if string
                    args
@@ -77,11 +79,11 @@
        (GO ,tag1)
        ,tag2)))
 ;; ----------------------------------------------------------------------------
-(defun typecase-error-string (keyform typelist)
+(defun typecase-error-string (keyform typelist) ; ABI
   (format nil
     (TEXT "The value of ~S must be of one of the types ~{~S~^, ~}")
     keyform typelist))
-(defun case-error-string (keyform caselist)
+(defun case-error-string (keyform caselist) ; ABI
   (format nil
     (TEXT "The value of ~S must be one of ~{~S~^, ~}")
     keyform caselist))
@@ -146,43 +148,50 @@
                                 ',place)
                         (SETF ,place (READ *QUERY-IO*))
                         (GO ,h)))))))))
-    (defmacro etypecase (keyform &rest keyclauselist)
+    (predefmacro etypecase (keyform &rest keyclauselist)
       (if (assoc t keyclauselist)
           `(typecase ,keyform ,@keyclauselist)
           (simply-error 'TYPECASE keyform keyclauselist
                         (typecase-errorstring keyform keyclauselist)
                         (typecase-expected-type keyclauselist))))
-    (defmacro ctypecase (keyplace &rest keyclauselist)
+    (predefmacro ctypecase (keyplace &rest keyclauselist)
       (if (assoc t keyclauselist)
           `(typecase ,keyplace ,@keyclauselist)
           (retry-loop 'TYPECASE keyplace keyclauselist
                       (typecase-errorstring keyplace keyclauselist))))
-    (defmacro ecase (keyform &rest keyclauselist)
+    (predefmacro ecase (keyform &rest keyclauselist)
       (simply-error 'CASE keyform keyclauselist
                     (case-errorstring keyform keyclauselist)
                     (case-expected-type keyclauselist)))
-    (defmacro ccase (keyform &rest keyclauselist)
+    (predefmacro ccase (keyform &rest keyclauselist)
       (retry-loop 'CASE keyform keyclauselist
                   (case-errorstring keyform keyclauselist)))))
 ;; ----------------------------------------------------------------------------
-(defmacro deftype (name lambdalist &body body)
+(defmacro deftype (&whole whole-form
+                   name lambdalist &body body)
   (unless (symbolp name)
     (error-of-type 'source-program-error
+      :form whole-form
+      :detail name
       (TEXT "type name should be a symbol, not ~S")
       name))
   (if (or (get name 'TYPE-SYMBOL) (get name 'TYPE-LIST))
     (error-of-type 'source-program-error
+      :form whole-form
+      :detail name
       (TEXT "~S is a built-in type and may not be redefined.")
       name))
   (multiple-value-bind (body-rest declarations docstring)
       (SYSTEM::PARSE-BODY body t)
     (if declarations (setq declarations (list (cons 'DECLARE declarations))))
-    (let ((%arg-count 0) (%min-args 0) (%restp nil)
+    (let ((%whole-form whole-form)
+          (%arg-count 0) (%min-args 0) (%restp nil) (%null-tests nil)
           (%let-list nil) (%keyword-tests nil) (%default-form '(QUOTE *)))
       (analyze1 lambdalist '(CDR <DEFTYPE-FORM>) name '<DEFTYPE-FORM>)
       (let ((lengthtest (make-length-test '<DEFTYPE-FORM>))
             (mainform `(LET* ,(nreverse %let-list)
                          ,@declarations
+                         ,@(nreverse %null-tests)
                          ,@(nreverse %keyword-tests)
                          (BLOCK ,name ,@body-rest))))
         (if lengthtest
@@ -197,30 +206,59 @@
                  (LAMBDA (<DEFTYPE-FORM>) ,mainform)))
              (SYS::%SET-DOCUMENTATION ',name 'TYPE ',docstring)
              ',name))))))
-(defun type-call-error (deftype-form)
+(defun type-call-error (deftype-form) ; ABI
   (error-of-type 'error
     (TEXT "The deftype expander for ~S may not be called with ~S arguments.")
     (car deftype-form) (1- (length deftype-form))))
 ;; ----------------------------------------------------------------------------
 ;; cf. X3J13 vote <173>
-(defmacro define-symbol-macro (symbol expansion)
+(defmacro define-symbol-macro (&whole whole-form
+                               symbol expansion)
   (unless (symbolp symbol)
     (error-of-type 'source-program-error
+      :form whole-form
+      :detail symbol
       (TEXT "~S: the name of a symbol macro must be a symbol, not ~S")
       'define-symbol-macro symbol))
   `(LET ()
      (EVAL-WHEN (COMPILE LOAD EVAL)
        (CHECK-NOT-SPECIAL-VARIABLE-P ',symbol)
-       (MAKUNBOUND ',symbol)
-       (SYSTEM::SET-SYMBOL-VALUE ',symbol
-                                 (SYSTEM::MAKE-SYMBOL-MACRO ',expansion)))
+       (SYSTEM::%PUT ',symbol 'SYSTEM::SYMBOLMACRO
+                     (SYSTEM::MAKE-GLOBAL-SYMBOL-MACRO ',expansion))
+       (SYSTEM::%PROCLAIM-SYMBOL-MACRO ',symbol))
      ',symbol))
-(defun check-not-special-variable-p (symbol)
+(defun check-not-special-variable-p (symbol) ; ABI
   (when (special-variable-p symbol)
     (error-of-type 'program-error
       (TEXT "~S: the symbol ~S names a global variable")
       'define-symbol-macro symbol)))
 ;; ----------------------------------------------------------------------------
+;; X3J13 vote <123>
+;; Macro (nth-value n form) == (nth n (multiple-value-list form)), CLtL2 S. 184
+(defmacro nth-value (n form)
+  (if (and (integerp n) (>= n 0))
+    (if (< n (1- multiple-values-limit))
+      (if (= n 0)
+        `(PROG1 ,form)
+        (let* ((resultvar (gensym))
+               (vars (list resultvar))
+               (ignores '()))
+          (do ((i n (1- i)))
+              ((zerop i))
+            (let ((g (gensym)))
+              (setq vars (cons g vars))
+              (setq ignores (cons g ignores))
+          ) )
+          `(MULTIPLE-VALUE-BIND ,vars ,form
+             (DECLARE (IGNORE ,@ignores))
+             ,resultvar
+           )
+      ) )
+      `(PROGN ,form NIL)
+    )
+    `(NTH ,n (MULTIPLE-VALUE-LIST ,form))
+) )
+;; ----------------------------------------------------------------------------
 (defun gensym-list (how-many)
   (map-into (make-list (if (numberp how-many) how-many (length how-many)))
             #'gensym))
@@ -268,7 +306,7 @@
        (UNWIND-PROTECT
          (PROGN ,@body-rest)
          ,@(if sindex
-               `((SETF ,index (SYSTEM::STRING-INPUT-STREAM-INDEX ,var))) '())
+             `((SETF ,index (SYSTEM::STRING-INPUT-STREAM-INDEX ,var))) '())
          (CLOSE ,var)))))
 ;; ----------------------------------------------------------------------------
 (defmacro with-open-file ((stream &rest options) &body body)
@@ -289,21 +327,31 @@
          (CLOSE ,var :ABORT T)))))
 ;; ----------------------------------------------------------------------------
 (defmacro with-output-to-string ((var &optional (string nil)
-                                      &key (element-type ''CHARACTER))
+                                  &key (element-type ''CHARACTER))
                                  &body body)
   (multiple-value-bind (body-rest declarations) (SYSTEM::PARSE-BODY body)
     (if string
-      `(LET ((,var (SYS::MAKE-STRING-PUSH-STREAM ,string)))
-         (DECLARE (READ-ONLY ,var) ,@declarations)
-         (UNWIND-PROTECT
-           (PROGN ,element-type ,@(or body-rest '(NIL)))
-           (CLOSE ,var)))
+      (let ((ignored-var (gensym)))
+        `(LET ((,var (SYS::MAKE-STRING-PUSH-STREAM ,string))
+               (,ignored-var ,element-type))
+           (DECLARE (READ-ONLY ,var) (IGNORE ,ignored-var) ,@declarations)
+           (UNWIND-PROTECT
+             (PROGN ,@body-rest)
+             (CLOSE ,var))))
       `(LET ((,var (MAKE-STRING-OUTPUT-STREAM :ELEMENT-TYPE ,element-type)))
-         (DECLARE ,@declarations)
+         (DECLARE (READ-ONLY ,var) ,@declarations)
          (UNWIND-PROTECT
            (PROGN ,@body-rest (GET-OUTPUT-STREAM-STRING ,var))
            (CLOSE ,var))))))
 ;; ----------------------------------------------------------------------------
+;; X3J13 vote <40>
+(defmacro print-unreadable-object
+          ((&whole args object stream &key type identity) &body body)
+  (declare (ignore object stream type identity))
+  `(SYSTEM::WRITE-UNREADABLE
+     ,(if body `(FUNCTION (LAMBDA () ,@body)) 'NIL)
+     ,@args))
+;; ----------------------------------------------------------------------------
 (in-package "EXT")
 (export '(space with-output-to-printer))
 (in-package "SYSTEM")
@@ -319,10 +367,10 @@
          (PROGN ,@body-rest)
          (CLOSE ,var)))))
 #+UNIX
-(defun make-printer-stream (&key (external-format :default))
+(defun make-printer-stream (&key (external-format :default)) ; ABI
   (make-pipe-output-stream "lpr" :external-format external-format))
 #+WIN32
-(defun make-printer-stream (&key (external-format :default))
+(defun make-printer-stream (&key (external-format :default)) ; ABI
   (open "prn" :direction :output :external-format external-format))
 ;; ----------------------------------------------------------------------------
 (in-package "EXT")
